const HOOK_NAME = Symbol('_sharding')
const ORIGINAL_TABLE_NAME = Symbol('_originalTableName')

// https://github.com/sequelize/sequelize/blob/master/lib/hooks.js
// 数据库分表中需要使用到的钩子，仅部分函数通过测试，如果有需要则新增钩子
const DEAFULT_TARGET_HOOKS = ['Save', 'Find', 'BulkUpdate', 'BulkDestroy', 'Upsert']
const DEAFULT_BEFORE_HOOKS = DEAFULT_TARGET_HOOKS.map(targetHook => 'before' + targetHook)
const DEAFULT_AFTER_HOOKS = DEAFULT_TARGET_HOOKS.map(targetHook => 'after' + targetHook)

const GENERATE_TABLE_NAME_BY_YYYYMM = Symbol('_generate_table_name_by_YYYYMM')

module.exports = {
    /**
     * 数据库分表，自动生成对应的表名字
     * @param sequelizeModel    需要分表的Model
     * @param generateTableName 生成表名字的函数，需要返回表名字，入参为instance，option，tableName，默认为 TABLENAME_YYYY_MM
     */
    dbSharding: function (sequelizeModel, generateTableName = this[GENERATE_TABLE_NAME_BY_YYYYMM]) {

        if (!sequelizeModel) throw new Error('Missing required arg: sequelizeModel')
        if (typeof generateTableName !== 'function') {
            throw new Error('generatePostfixFunc is not a function')
        }

        let beforeHooks = DEAFULT_BEFORE_HOOKS
        let afterHooks = DEAFULT_AFTER_HOOKS

        // 在数据库操作钩子前修改要操作的数据表名
        beforeHooks.forEach(hookName => {
            sequelizeModel.removeHook(hookName, HOOK_NAME)
            if (!sequelizeModel[ORIGINAL_TABLE_NAME]) {
                sequelizeModel[ORIGINAL_TABLE_NAME] = sequelizeModel.tableName
            }
            sequelizeModel.addHook(hookName, HOOK_NAME, (instance, options) => {
                sequelizeModel.tableName = generateTableName(instance, options, sequelizeModel[ORIGINAL_TABLE_NAME])
            })
        })
        // 在数据库操作钩子前恢复回原来的数据表名
        afterHooks.forEach(hookName => {
            sequelizeModel.removeHook(hookName, HOOK_NAME)
            sequelizeModel.addHook(hookName, HOOK_NAME, (instance, options) => {
                if (sequelizeModel[ORIGINAL_TABLE_NAME]) {
                    sequelizeModel.tableName = sequelizeModel[ORIGINAL_TABLE_NAME]
                }
            })

        })
    },

    // 生成YYYYMM的数据库表明
    [GENERATE_TABLE_NAME_BY_YYYYMM](instance, options, tableName) {
        const date = new Date()
        let yyyymm = date.getFullYear() + '_' + ('0' + (date.getMonth() + 1)).slice(-2)
        return `${tableName}_${yyyymm}`
    }
}
